@@ -34,7 +34,15 @@ module System.OsString.MODULE_NAME
   , encodeWith
   , encodeFS
   , encodeLE
+#ifdef WINDOWS
+  , fromString
+#endif
   , fromBytes
+  , fromShortBytes
+#ifndef WINDOWS
+  , fromBytestring
+  , fromShortBytestring
+#endif
   , pstr
   , singleton
   , empty
@@ -137,7 +145,7 @@ module System.OsString.MODULE_NAME
 where
 
 
-
+import System.OsString.Internal.Exception
 import System.OsString.Internal.Types (
 #ifdef WINDOWS
   WindowsString(..), WindowsChar(..)
@@ -152,30 +160,47 @@ import Control.Monad.Catch
     ( MonadThrow, throwM )
 import Data.ByteString.Internal
     ( ByteString )
+import Data.ByteString.Short.Internal
+    ( ShortByteString )
 import Control.Exception
     ( SomeException, try, displayException )
 import Control.DeepSeq ( force )
 import Data.Bifunctor ( first )
-import GHC.IO
-    ( evaluate, unsafePerformIO )
+import Control.Exception ( evaluate )
+import System.IO.Unsafe ( unsafePerformIO )
 import qualified GHC.Foreign as GHC
+#if __GLASGOW_HASKELL__ >= 914
+import Language.Haskell.TH.Lift
+    (Lift(..))
+import Language.Haskell.TH.QuasiQuoter
+    ( QuasiQuoter (..) )
+#else
+import Language.Haskell.TH.Syntax
+    (Lift(..))
 import Language.Haskell.TH.Quote
     ( QuasiQuoter (..) )
-import Language.Haskell.TH.Syntax
-    ( Lift (..), lift )
+#endif
 
 
 import GHC.IO.Encoding.Failure ( CodingFailureMode(..) )
 #ifdef WINDOWS
 import System.OsString.Encoding
+#if defined(__MHS__)
+import GHC.IO.Encoding( TextEncoding, utf16le )
+#else
 import System.IO
     ( TextEncoding, utf16le )
+#endif
 import GHC.IO.Encoding.UTF16 ( mkUTF16le )
 import qualified System.OsString.Data.ByteString.Short.Word16 as BSP
 #else
 import System.OsString.Encoding
+#if defined(__MHS__)
+import GHC.IO.Encoding( TextEncoding, utf8 )
+#else
 import System.IO
     ( TextEncoding, utf8 )
+#endif
 import GHC.IO.Encoding.UTF8 ( mkUTF8 )
 import qualified System.OsString.Data.ByteString.Short as BSP
 #endif
@@ -230,7 +255,7 @@ encodeWith :: TextEncoding  -- ^ text encoding (wide char)
            -> String
            -> Either EncodingException PLATFORM_STRING
 encodeWith enc str = unsafePerformIO $ do
-  r <- try @SomeException $ GHC.withCStringLen enc str $ \cstr -> WindowsString <$> BS8.packCStringLen cstr
+  r <- trySafe @SomeException $ GHC.withCStringLen enc str $ \cstr -> WindowsString <$> BS8.packCStringLen cstr
   evaluate $ force $ first (flip EncodingError Nothing . displayException) r
 #else
 -- | Encode a 'String' with the specified encoding.
@@ -238,7 +263,7 @@ encodeWith :: TextEncoding
            -> String
            -> Either EncodingException PLATFORM_STRING
 encodeWith enc str = unsafePerformIO $ do
-  r <- try @SomeException $ GHC.withCStringLen enc str $ \cstr -> PosixString <$> BSP.packCStringLen cstr
+  r <- trySafe @SomeException $ GHC.withCStringLen enc str $ \cstr -> PosixString <$> BSP.packCStringLen cstr
   evaluate $ force $ first (flip EncodingError Nothing . displayException) r
 #endif
 
@@ -289,6 +314,21 @@ encodeLE = fmap WindowsString . encodeWithBaseWindows
 encodeLE = fmap PosixString . encodeWithBasePosix'
 #endif
 
+#ifdef WINDOWS
+-- | Like 'encodeLE but not in IO.
+--
+-- 'encodeLE' was designed to have a symmetric type signature
+-- on unix and windows, but morally the function has no IO effects on windows,
+-- so we provide this variant without breaking existing API.
+--
+-- On windows, 'encodeLE' is equivalent to 'encodeFS'.
+--
+-- This function does not exist on unix.
+--
+-- @since 2.0.6
+fromString :: String -> WindowsString
+fromString = unsafePerformIO . fmap WindowsString . encodeWithBaseWindows
+#endif
 
 #ifdef WINDOWS_DOC
 -- | Partial unicode friendly decoding.
@@ -319,7 +359,7 @@ decodeWith :: TextEncoding
            -> PLATFORM_STRING
            -> Either EncodingException String
 decodeWith winEnc (WindowsString ba) = unsafePerformIO $ do
-  r <- try @SomeException $ BS8.useAsCStringLen ba $ \fp -> GHC.peekCStringLen winEnc fp
+  r <- trySafe @SomeException $ BS8.useAsCStringLen ba $ \fp -> GHC.peekCStringLen winEnc fp
   evaluate $ force $ first (flip EncodingError Nothing . displayException) r
 #else
 -- | Decode a 'PosixString' with the specified encoding.
@@ -329,7 +369,7 @@ decodeWith :: TextEncoding
        -> PLATFORM_STRING
        -> Either EncodingException String
 decodeWith unixEnc (PosixString ba) = unsafePerformIO $ do
-  r <- try @SomeException $ BSP.useAsCStringLen ba $ \fp -> GHC.peekCStringLen unixEnc fp
+  r <- trySafe @SomeException $ BSP.useAsCStringLen ba $ \fp -> GHC.peekCStringLen unixEnc fp
   evaluate $ force $ first (flip EncodingError Nothing . displayException) r
 #endif
 
@@ -397,12 +437,53 @@ decodeFS (PosixString ba) = decodeWithBasePosix ba
 fromBytes :: MonadThrow m
           => ByteString
           -> m PLATFORM_STRING
+fromBytes = fromShortBytes . BS16.toShort
+
+#ifdef WINDOWS_DOC
+-- | Constructs a platform string from a ShortByteString.
+--
+-- This ensures valid UCS-2LE.
+-- Note that this doesn't expand Word8 to Word16 on windows, so you may get invalid UTF-16.
+--
+-- Throws 'EncodingException' on invalid UCS-2LE (although unlikely).
+--
+-- @since 2.0.8
+#else
+-- | Constructs a platform string from a ShortByteString.
+--
+-- This is a no-op.
+--
+-- @since 2.0.8
+#endif
+fromShortBytes :: MonadThrow m
+               => ShortByteString
+               -> m PLATFORM_STRING
 #ifdef WINDOWS
-fromBytes bs =
-  let ws = WindowsString . BS16.toShort $ bs
+fromShortBytes bs =
+  let ws = WindowsString bs
   in either throwM (const . pure $ ws) $ decodeWith ucs2le ws
 #else
-fromBytes = pure . PosixString . BSP.toShort
+fromShortBytes = pure . PosixString
+#endif
+
+#ifndef WINDOWS
+-- | Like 'fromBytes', but not in IO.
+--
+-- 'fromBytes' was designed to have a symmetric type signature
+-- on unix and windows, but morally the function has no IO effects on unix,
+-- so we provide this variant without breaking existing API.
+--
+-- This function does not exist on windows.
+--
+-- @since 2.0.6
+fromBytestring :: ByteString -> PosixString
+fromBytestring = PosixString . BSP.toShort
+
+-- | Like 'fromShortBytes', but not in IO, similarly to 'fromBytestring'
+--
+-- @since 2.0.8
+fromShortBytestring :: ShortByteString -> PosixString
+fromShortBytestring = PosixString
 #endif
 
 
@@ -413,6 +494,7 @@ fromBytes = pure . PosixString . BSP.toShort
 -- | QuasiQuote a 'PosixString'. This accepts Unicode characters
 -- and encodes as UTF-8 on unix.
 #endif
+#if defined(MIN_VERSION_template_haskell) || defined(MIN_VERSION_template_haskell_quasi_quoter)
 pstr :: QuasiQuoter
 pstr =
   QuasiQuoter
@@ -441,7 +523,10 @@ pstr =
       fail "illegal QuasiQuote (allowed as expression or pattern only, used as a declaration)"
   }
 #endif
-
+#else
+pstr :: a
+pstr = error "Systen.OsString.Common.pstr: no Template Haskell"
+#endif /* !defined(__MHS__) */
 
 -- | Unpack a platform string to a list of platform words.
 unpack :: PLATFORM_STRING -> [PLATFORM_WORD]
